#
#	 AppDelegate.rb
#	 LogNoter
#
#	 Created by Gregory Clarke on 5/2/16.
#	 Copyright (c) 2016 Intelligent Assistance. All rights reserved.
#
# - ERB::Util.url_encode DONE


require 'osx/cocoa'
include OSX
require 'pp'
require 'erb'
bundle = "#{File.dirname(__FILE__)}/noteLogger_methods.bundle"
if File.exist?(bundle)
	require 'noteLogger_methods.bundle'
else
	p "NO BUNDLE #{bundle} !!!\n"
end

class AppDelegate < OSX::NSObject
	ib_outlets :debugMenu, :localhostMenuItem
	ib_outlets :window, :loginButton, :eventPopup, :columnTypePopup, :keywordTypePopup, :logNoteButton, :spinner, :uploadButton, :offsetCheckbox, :offsetDatePicker, :locationsTableView, :peopleTableView, :activitiesTableView, :othersTableView, :notesScrollView, :notesTableView, :keywordComboBox, :leadingConstraint
	ib_outlets :preferencesWindow, :preferencesTabs, :usernameField, :passwordField, :refreshRadioButtons

	def initialize
		default_values = { "ColumnType" => "All", "KeywordType" => "Other", "Localhost" => "0", "Debug" => 0, "Refresh" => "0", "ShowKeywordsColumns" => 1 }
		defaultsController = NSUserDefaultsController.sharedUserDefaultsController
		defaultsController.setInitialValues(default_values)
	end

	def awakeFromNib
		# get and set @notesTableView leading constraint (using AutoLayout to get and set the constant for the leading of notes table
		@no_keyword_columns_leading = 20.0 # shows only the Notes table
		@one_keyword_column_leading = 208.0 # shows Locations table and Notes table
		@all_keyword_columns_leading = 772.0 # shows 4 Keywords tables and Notes table
		puts "@leadingConstraint.constant: #{@leadingConstraint.constant}"
		# defaults = NSUserDefaults.standardUserDefaults
		# if defaults.objectForKey("ShowKeywordsColumns").nil?
		# elsif defaults.objectForKey("ShowKeywordsColumns")==1
		# 	@leadingConstraint.constant = @all_keyword_columns_leading
		# else
		# 	@leadingConstraint.constant = @no_keyword_columns_leading
		# end
		chooseKeywordColumns(self)
		unlockInterface(false)
		@location_keywords, @people_keywords, @activity_keywords, @other_keywords = Array.new, Array.new, Array.new, Array.new
		@metadata_logs = Array.new
		@notes_metadata_logs, @autocomplete_keywords = Array.new, Array.new
		@serverTimestampFormatter = NSDateFormatter.alloc.init
		enUSPOSIXLocale = NSLocale.localeWithLocaleIdentifier("en_US_POSIX")
		@serverTimestampFormatter.setLocale(enUSPOSIXLocale)
		@serverTimestampFormatter.setDateFormat("yyyy-MM-dd HH:mm:ss")
		@serverTimestampFormatter.setTimeZone(NSTimeZone.timeZoneWithName("America/Los_Angeles"))
		# zero_to_four = NSMakeRange(0, 4) # [0,1,2,3] FIXME: causes crash on Big SUr
		# @fourColumnIndexes = NSIndexSet.indexSetWithIndexesInRange(zero_to_four)
		@offsetSeconds = 0
		@offsetDatePicker.dateValue = NSDate.date # set to current date/time
		### check for Application Support subfolder
		paths = NSFileManager.defaultManager.URLsForDirectory_inDomains(NSApplicationSupportDirectory, NSUserDomainMask)
		if paths.length > 0
			app_support_subfolder = paths[0].path.stringByExpandingTildeInPath + "/com.Lumberjack.noteLogger"
			Dir.mkdir(app_support_subfolder) unless File.directory?(app_support_subfolder)
			@marshal_will_and_holly = app_support_subfolder + "/data.ljn"
			@argonaut = app_support_subfolder + "/data.json"
		end
		@local_logs_db = Array.new
		if File.exist?(@marshal_will_and_holly)
			file = File.open(@marshal_will_and_holly, 'r')
			begin
				@local_logs_db = Marshal.load file.read
			rescue
				puts "UNABLE TO MARSHAL LOAD!!!!"
			end
			file.close
			# puts "Marshal: @local_logs_db = #{@local_logs_db.inspect}", "---"
		end
		if File.exist?(@argonaut)
			file = NSInputStream.alloc.initWithFileAtPath(@argonaut)
			file.open
			begin
				@local_logs_db = NSJSONSerialization.JSONObjectWithStream_options_error(file, 0, nil).to_ruby
			rescue
				puts "UNABLE TO JSON PARSE!!!!"
			end
			file.close
			# puts "JSON: @local_logs_db = #{@local_logs_db.inspect}", "---"
		end
	end

	def applicationWillFinishLaunching(sender)
		LetsMove.PFMoveToApplicationsFolderIfNecessary
	end

	def applicationDidFinishLaunching(notification)
		@logged_in = false
		getCredentialsFromKeychain()
		get_uris()
		if $localhost # can't be in awakeFromNib
			alert = NSAlert.alloc.init
			alert.setMessageText("Debug has been enabled:")
			alert.setInformativeText("localhost = #{$localhost}\n")
			alert.setAlertStyle(NSCriticalAlertStyle)
			alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
		end
		login(self)
		@can_add_logs = false
		@logNoteButton.setEnabled(false)
		@timer = nil
		clickRefreshRadioButtons(self) # add NSTimer if necessary
	end
	
	def clickRefreshRadioButtons(sender)
		unless @refreshRadioButtons.selectedCell.nil?
			refresh_interval = @refreshRadioButtons.selectedCell.tag.to_f
			puts @timer.inspect, refresh_interval
			if refresh_interval==0.0 && @timer.nil?
				puts "don't add @timer"
			elsif refresh_interval==0.0 && @timer!=nil
				puts "invalidate @timer"
				@timer.invalidate # @timer will be nil
			else
				puts "create and add @timer every #{refresh_interval} seconds"
				@timer = NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats(refresh_interval, self, 'timerFired:', nil, true)
				run_loop = NSRunLoop.currentRunLoop
				run_loop.addTimer_forMode(@timer, NSDefaultRunLoopMode)
				# run_loop.addTimer_forMode(@timer, NSEventTrackingRunLoopMode)
			end
			canAddLogs()
		end
	end
	ib_action :clickRefreshRadioButtons
	
	def canAddLogs()
		puts "canAddLogs", @refreshRadioButtons.selectedCell.tag.inspect
		if @eventPopup.indexOfSelectedItem==0
			@can_add_logs = false
		elsif @refreshRadioButtons.selectedCell.tag==0
			@can_add_logs = true
		else
			@can_add_logs = false
			@uploadButton.setEnabled(false)
		end
		puts "@can_add_logs = #{@can_add_logs}"
		@logNoteButton.setEnabled(@can_add_logs)
	end

	def showPreferences(sender)
		getCredentialsFromKeychain() # in case they've been changed in Keychain Access
		@preferencesWindow.makeKeyAndOrderFront(self)
	end
	ib_action :showPreferences

	def showHelp(sender)
		system('open', "http://lumberjacksystem.com/notelogger.html")
	end
	ib_action :showHelp
	
	def emailSupport(sender)
    executable = NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleExecutable")
		version = NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleShortVersionString")
		app = "#{executable} (version #{version})".to_ns.stringByAddingPercentEscapesUsingEncoding(4)
		body = "What went wrong?\n\n\n".to_ns.stringByAddingPercentEscapesUsingEncoding(4)
		mailto = NSURL.URLWithString("mailto:info@LumberjackSystem.com?subject=Bug%20Report%20for%20#{app}&body=#{body}")
		result = NSWorkspace.sharedWorkspace.openURL(mailto)
		if result==false
			contactus = NSURL.URLWithString("http://lumberjacksystem.com/about.html")
			NSWorkspace.sharedWorkspace.openURL(contactus)
		end
	end
	ib_action :emailSupport

	def showGlossary(sender)
		system('open', "http://lumberjacksystem.com/glossary.html")
	end
	ib_action :showGlossary

	def login(sender) # FIXME: weird log in error message when no internet
		if @logged_in
			clickUploadButton(self)
			begin
				if @logout.nil? # haven't gotten the URIs yet
					get_uris()
				end
				response = get_NSMutableURLRequest_response(@logout)
				#puts response.inspect
				lock = NSImage.imageNamed("NSLockLockedTemplate")
				@loginButton.setImage(lock)
				@logged_in = false
				unlockInterface(false)
			rescue => e # no net connection, can't find server...
				#puts "...problem getting the response"
				@spinner.stopAnimation(self)
				alert = NSAlert.alloc.init
				alert.setMessageText("There was a problem signing in to the Lumberjack server")
				alert.setInformativeText(e.message)
				alert.setAlertStyle(NSCriticalAlertStyle)
				alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
			end
		else
			get_uris()
			@spinner.startAnimation(self)
			puts "start to connect... localhost? #{$localhost}... #{@login}"
			begin
				username = ERB::Util.url_encode(@usernameField.stringValue)
				password = ERB::Util.url_encode(@passwordField.stringValue)
				post = "username=#{username}&password=#{password}&lumberyard=0"
				json_response = get_NSMutableURLRequest_response(@login, post)
				unless json_response.empty? # no username or password, so ignore
					if json_response["customer_id"]=="0"
						raise json_response["message"]
					end
					message = json_response["message"]
					@user_id = json_response["user_id"]
					@customer_id = json_response["customer_id"]
					#puts "@user_id = #{@user_id}, @customer_id = #{@customer_id}, #{message}"
					get_uris()
					response = get_NSMutableURLRequest_response(@list_stories)
					puts "response: #{response.inspect}"
					if @user_id=="0"
						raise message
					else
						lock = NSImage.imageNamed("NSLockUnlockedTemplate")
						@loginButton.setImage(lock)
						@logged_in = true
						unlockInterface(true)
						# get all the stories, create popup menu
						@stories = [ {"event" => ""} ] + response # hold this result just in case a story is added to the database later
						@eventPopup.removeAllItems
						@stories.each_with_index do |s, i|
							@eventPopup.addItemWithTitle(s["event"].to_s) # to_s fixes empty event which returns as NULL
							# @eventPopup.menu.addItemWithTitle_action_keyEquivalent(s["event"], nil, "") # add items to the menu instead of the popup button so that duplicate event names can be supported
							unless i==0
								has_event = @local_logs_db.find { |m| m["story_id"]==s["id"] }
								if has_event.nil?
									@local_logs_db << { "story_id" => s["id"], "notes" => Array.new, "locations" => Array.new, "people" => Array.new, "activities" => Array.new, "others" => Array.new }
								end
							end
						end
						serialize_local_logs()
						#pp @local_logs_db
						chooseEvent(self)
						@report = "--------------------------------------------\n"
					end
				end
			rescue => e # no net connection, can't find server...
				puts "...problem getting the response"
				puts e.message
				message = e.message
				message = "Is your Mac connected to the internet?" if message=="NSInvalidArgumentException - data parameter is nil"
				@spinner.stopAnimation(self)
				alert = NSAlert.alloc.init
				alert.setMessageText("There was a problem signing in to the Lumberjack server")
				alert.setInformativeText(message)
				alert.setAlertStyle(NSCriticalAlertStyle)
				alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
			end
			@spinner.stopAnimation(self)
		end
	end
	ib_action :login

	def unlockInterface(bool)
		@eventPopup.setEnabled(bool)
		@columnTypePopup.setEnabled(bool)
		@keywordTypePopup.setEnabled(bool)
		@logNoteButton.setEnabled(bool)
		@locationsTableView.setEnabled(bool)
		@peopleTableView.setEnabled(bool)
		@activitiesTableView.setEnabled(bool)
		@othersTableView.setEnabled(bool)
		@notesTableView.setEnabled(bool)
	end

	# NOTE: getCredentialsFromKeychain/saveCredentialsToKeychain copied from Lumberyard
	def getCredentialsFromKeychain
		# puts "getCredentialsFromKeychain"
		username = ""
		password = ""
		keychain = `security find-generic-password -s 'Lumberjack account details' -g 2>&1` # redirect STDERR to STDOUT
		parts = keychain.split("\n")
		acct_string = parts.find { |p| p.include?("acct") }
		pssw_string = parts.find { |p| p.include?("password:") }
		if acct_string.nil? || pssw_string.nil?
			@usernameField.setStringValue('')
			@passwordField.setStringValue('')
			# log the user out if logged in
			@logged_in = true
			login(self)
		else
			splitAcct = acct_string.split('"') # split on "
			if splitAcct.length > 3
				username = splitAcct.last
			end
			splitPw = pssw_string.split('"')
			if splitPw.length > 1
				password = splitPw.last
			end
			@usernameField.setStringValue(username)
			@passwordField.setStringValue(password)
		end
		if username.empty? || password.empty?
			@preferencesWindow.makeKeyAndOrderFront(self)
			@preferencesTabs.selectFirstTabViewItem(self)
		end
	end

	def saveCredentialsToKeychain(sender)
		username = @usernameField.stringValue.strip
		if username.include?("mailto:")
			username = username.sub("mailto:", "")
		end
		@usernameField.stringValue = username
		password = @passwordField.stringValue.strip
		@passwordField.stringValue = password
		`security delete-generic-password -s 'Lumberjack account details'`
		`security add-generic-password -a '#{username}' -s 'Lumberjack account details' -w '#{password}'`
		@preferencesWindow.orderOut(self)
	end
	ib_action :saveCredentialsToKeychain

	def chooseEvent(sender)
		puts "chooseEvent"
		canAddLogs()
		if @eventPopup.indexOfSelectedItem==0
			@location_keywords, @people_keywords, @activity_keywords, @other_keywords = Array.new, Array.new, Array.new, Array.new
			@notes_metadata_logs = Array.new
			@locationsTableView.reloadData
			@peopleTableView.reloadData
			@activitiesTableView.reloadData
			@othersTableView.reloadData
			@notesTableView.reloadData
		else
			story = @stories[@eventPopup.indexOfSelectedItem]
			@not_yet_uploaded = @local_logs_db.find { |m| m["story_id"]==story["id"] }
			#puts "@not_yet_uploaded = #{@not_yet_uploaded.inspect}"
			@not_yet_uploaded["locations"] = [] if @not_yet_uploaded["locations"].nil?
			@not_yet_uploaded["people"] = [] if @not_yet_uploaded["people"].nil?
			@not_yet_uploaded["activities"] = [] if @not_yet_uploaded["activities"].nil?
			@not_yet_uploaded["others"] = [] if @not_yet_uploaded["others"].nil?
			local_locations_logs, local_people_logs, local_activities_logs, local_others_logs, local_notes_logs = Array.new, Array.new, Array.new, Array.new, Array.new
			unless @not_yet_uploaded.nil?
				local_locations_logs = @not_yet_uploaded["locations"]
				local_people_logs = @not_yet_uploaded["people"]
				local_activities_logs = @not_yet_uploaded["activities"]
				local_others_logs = @not_yet_uploaded["others"]
				local_notes_logs = @not_yet_uploaded["notes"]
				if local_locations_logs.length > 0 || local_people_logs.length > 0 || local_activities_logs.length > 0 || local_others_logs.length > 0 || local_notes_logs.length > 0
					@uploadButton.setEnabled(true)
				end
			end
			get_server_keywords_and_logs(story, local_notes_logs)
		end
	end
	ib_action :chooseEvent
	
	def clickTimeOffsetCheckbox(sender)
		if @offsetCheckbox.state==0
			@offsetSeconds = 0
		else
			offsetTime = @offsetDatePicker.dateValue
			delta = offsetTime.timeIntervalSinceNow
			@offsetSeconds = delta
		end
		puts "@offsetSeconds = #{@offsetSeconds}"
	end
	ib_action :clickTimeOffsetCheckbox
	
	def timerFired(sender)
		puts "timerFired"
		story = @stories[@eventPopup.indexOfSelectedItem]
		get_server_keywords_and_logs(story, [])
	end
	
	def get_server_keywords_and_logs(story, local_notes_logs)
		if @eventPopup.indexOfSelectedItem!=0
			@spinner.startAnimation(self)
			@location_keywords = get_NSMutableURLRequest_response("#{@list_locations}/#{story["id"]}")
			@people_keywords = get_NSMutableURLRequest_response("#{@list_people}/#{story["id"]}")
			@activity_keywords = get_NSMutableURLRequest_response("#{@list_activities}/#{story["id"]}")
			@other_keywords = get_NSMutableURLRequest_response("#{@list_keywords}/#{story["id"]}")
			unsorted_metadata_logs = local_notes_logs + get_NSMutableURLRequest_response("#{@list_logs_and_notes}/#{story["id"]}")
			#puts "#{unsorted_metadata_logs.count} log rows"
			@notes_metadata_logs, @logged_keywords = Array.new, Array.new
			unsorted_metadata_logs.sort_by { |m| m["start_time"] }.each do |m|
				# convert metadata into keyword for tableview
				keyword = "#{m["metadata"][2..-1]}"
				case m["metadata"][0,2]
				when "p:"
					keyword << " - Person"
				when "a:"
					keyword << " - Activity"
				when "k:"
					keyword << " - Other"
				when "no" # notes logged from Comments field
					keyword = "Comment"
				when "fa" # Favorite
					keyword = "Favorite"
				when "Ma" # Marker
					keyword = "Marker"
				end
				m["keyword"] = keyword
				# convert Los Angeles timestamp strings (with no time zone) to UTC Date/Time - server time is set to be LA instead of just using Time.parse which uses computer's local time zone
				start_time = @serverTimestampFormatter.dateFromString(m["start_time"])
				localizedTime = NSDateFormatter.localizedStringFromDate_dateStyle_timeStyle(start_time, NSDateFormatterShortStyle, NSDateFormatterMediumStyle)
				#puts start_time, localizedTime
				m["local_start_time"] = localizedTime
				if m["stop_time"].nil? || m["stop_time"]=="0000-00-00 00:00:00" || m["stop_time"].class==NSNull # forgot to stop the logging
					m["local_stop_time"] = ""
				else
					stop_time = @serverTimestampFormatter.dateFromString(m["stop_time"])
					localizedTime = NSDateFormatter.localizedStringFromDate_dateStyle_timeStyle(stop_time, NSDateFormatterShortStyle, NSDateFormatterMediumStyle)
					m["local_stop_time"] = localizedTime
				end
				@notes_metadata_logs << m
				if keyword!="Comment" && keyword!="Favorite" && keyword!="Marker"
					@logged_keywords << keyword
				end
			end
			chooseKeywordType(self) # will build autocomplete_keywords and rebuildComboBox
			@locationsTableView.reloadData
			@peopleTableView.reloadData
			@activitiesTableView.reloadData
			@othersTableView.reloadData
			# sortDescriptor = @notesTableView.sortDescriptors.first
			# if sortDescriptor.ascending==false
			# 	@notes_metadata_logs = @notes_metadata_logs.reverse
			# end
			@notesTableView.reloadData
			unless @notes_metadata_logs.empty?
				@notesTableView.scrollRowToVisible(@notes_metadata_logs.length - 1)
			end
			@window.makeFirstResponder(@notesTableView) # blue focus ring around table
			@spinner.stopAnimation(self)
		end
	end
	
	def clickChangeViewButton(sender)
		defaults = NSUserDefaults.standardUserDefaults
		if @leadingConstraint.constant==@no_keyword_columns_leading
			@leadingConstraint.constant = @all_keyword_columns_leading
			defaults.setObject_forKey(true, "ShowKeywordsColumns")
		else
			@leadingConstraint.constant = @no_keyword_columns_leading
			defaults.setObject_forKey(false, "ShowKeywordsColumns")
		end
		puts defaults.objectForKey("ShowKeywordsColumns")
	end
	ib_action :clickChangeViewButton

	def chooseKeywordColumns(sender) # triggered when changing the visible columns
		pref = NSUserDefaults.standardUserDefaults.objectForKey("ColumnType")
		puts "chooseKeywordColumns: #{pref}"
		firstTableColumn = @locationsTableView.tableColumns[0]
		case @columnTypePopup.titleOfSelectedItem
		when "All"
			firstTableColumn.headerCell.stringValue = "Locations"
			@leadingConstraint.constant = @all_keyword_columns_leading
		when "None"
			firstTableColumn.headerCell.stringValue = "Locations"
			@leadingConstraint.constant = @no_keyword_columns_leading
		when "Locations"
			firstTableColumn.headerCell.stringValue = "Locations"
			@leadingConstraint.constant = @one_keyword_column_leading
		when "People"
			firstTableColumn.headerCell.stringValue = "People"
			@leadingConstraint.constant = @one_keyword_column_leading
		when "Activities"
			firstTableColumn.headerCell.stringValue = "Activities"
			@leadingConstraint.constant = @one_keyword_column_leading
		when "Other"
			firstTableColumn.headerCell.stringValue = "Other Keywords"
			@leadingConstraint.constant = @one_keyword_column_leading
		end
		@locationsTableView.reloadData
	end
	ib_action :chooseKeywordColumns


	def chooseKeywordType(sender) # triggered when changing Events or changing keyword type
		puts "chooseKeywordType"
		if @eventPopup.indexOfSelectedItem!=0
			story = @stories[@eventPopup.indexOfSelectedItem]
			@event_keywords, event_keywords = Array.new, Array.new
			@keyword_prefix = ""
			case @keywordTypePopup.titleOfSelectedItem
			when "People"
				event_keywords = get_NSMutableURLRequest_response("#{@list_people}/#{story["id"]}")
				@keyword_prefix = "p:"
				#puts "People: #{event_keywords.inspect}"
			when "Activity"
				event_keywords = get_NSMutableURLRequest_response("#{@list_activities}/#{story["id"]}")
				@keyword_prefix = "a:"
				#puts "Activity: #{event_keywords.inspect}"
			when "Other"
				event_keywords = get_NSMutableURLRequest_response("#{@list_keywords}/#{story["id"]}")
				@keyword_prefix = "k:"
				#puts "Other: #{event_keywords.inspect}"
			end
			event_keywords.each { |e| @event_keywords << e["keyword"] }
			rebuildComboBox()
		end
	end
	ib_action :chooseKeywordType
	
	# FIXME: chooseCheckboxListKeywords popup menu populates the checkboxes in the left tableview

	def rebuildComboBox()
		autocomplete_keywords = (@logged_keywords + @event_keywords).uniq.sort
		#puts "rebuildComboBox #{autocomplete_keywords.inspect}"
		@keywordComboBox.removeAllItems
		@keywordComboBox.addItemsWithObjectValues( autocomplete_keywords) # add unique @autocomplete_keywords to the combo box
	end

	def clickLogButton(sender) # or keyboard equivalent
		puts "clickLogButton"
		if @can_add_logs
			timeNow = NSDate.date.dateByAddingTimeInterval(@offsetSeconds)
			serverTimestamp = @serverTimestampFormatter.stringFromDate(timeNow)
			localizedTimeString = NSDateFormatter.localizedStringFromDate_dateStyle_timeStyle(timeNow, NSDateFormatterShortStyle, NSDateFormatterMediumStyle)
			#puts serverTimestamp, localizedTimeString
			if @logNoteButton.title=="Start Log" # Start Log: add a new row to @notes_metadata_logs and the TableView
				@spinner.startAnimation(self)
				row_number = @notes_metadata_logs.length
				story = @stories[@eventPopup.indexOfSelectedItem]
				if @notes_metadata_logs.last!=nil && @notes_metadata_logs.last["stop_time"]==serverTimestamp
					timeNowPlus = timeNow.dateByAddingTimeInterval(1) # start this log 1 second after the previous stop time
					serverTimestamp = @serverTimestampFormatter.stringFromDate(timeNowPlus)
					localizedTimeString = NSDateFormatter.localizedStringFromDate_dateStyle_timeStyle(timeNowPlus, NSDateFormatterShortStyle, NSDateFormatterMediumStyle)
					puts "LOGS WILL OVERLAP", serverTimestamp, localizedTimeString
				end
				case @keywordTypePopup.titleOfSelectedItem
				when "Marker"
					keyword = "Marker"
					metadata = "Marker"
				when "Favorite"
					keyword = "Favorite" # the column display
					metadata = "favorite" # the metadata_logs entry
				else
					keyword = ""
					metadata = "" # will get set by tableview editing
				end
				new_log_row = { "id" => "", "story_id" => story["id"], "card" => @user_id, "start_time" => serverTimestamp, "local_start_time" => localizedTimeString, "metadata" => metadata, "keyword" => keyword, "text" => "" }
				@notes_metadata_logs << new_log_row
				@logNoteButton.setTitle("Stop Log")
				@notesTableView.reloadData # will reload all @keywordComboBox in table view
				selectedRowIndex = NSIndexSet.indexSetWithIndex(row_number)
				@notesTableView.selectRowIndexes_byExtendingSelection(selectedRowIndex, false) # select the new added row
				# find the first editable column
				keywordColumn = 1
				notesColumn = 2
				@notesTableView.tableColumns.each_with_index do |tableColumn, c|
					if tableColumn.headerCell.stringValue=="Keyword"
						keywordColumn = c
					elsif tableColumn.headerCell.stringValue=="Notes"
						notesColumn = c
					end
				end
				if @keywordTypePopup.titleOfSelectedItem=="Marker" || @keywordTypePopup.titleOfSelectedItem=="Favorite"
					@notesTableView.editColumn_row_withEvent_select(notesColumn, row_number, nil, true) # begin editing the Note column of this row (and scroll to make the row visible)
				else # begin editing the first editable column of this row (and scroll to make the row visible)
					if keywordColumn < notesColumn
						@notesTableView.editColumn_row_withEvent_select(keywordColumn, row_number, nil, true) 
					else
						@notesTableView.editColumn_row_withEvent_select(notesColumn, row_number, nil, true) 
					end
				end
			else # Stop Log: send to database, get row id number
				#puts "Stop Log" # FIXME: rethink finding row without Stop Time
				#puts @notes_metadata_logs.last["metadata"].nil?
				#puts @notes_metadata_logs.last["metadata"].empty?
				unless @notes_metadata_logs.last["metadata"].nil? || @notes_metadata_logs.last["metadata"].empty?
					@spinner.stopAnimation(self)
					@logNoteButton.setTitle("Start Log")
					if @keywordTypePopup.titleOfSelectedItem=="Marker"
						@notes_metadata_logs.last["stop_time"] = @notes_metadata_logs.last["start_time"]
						@notes_metadata_logs.last["local_stop_time"] = @notes_metadata_logs.last["local_start_time"]
					else
						@notes_metadata_logs.last["stop_time"] = serverTimestamp
						@notes_metadata_logs.last["local_stop_time"] = localizedTimeString
					end
					# selectedRowIndex = NSIndexSet.indexSetWithIndex(@notesTableView.selectedRow)
					# @notesTableView.reloadDataForRowIndexes_columnIndexes(selectedRowIndex, @fourColumnIndexes) # don't need to reload data because it's already there
					@notesTableView.deselectAll(self)
					notes_row_id = sendRowToMySQL(@notes_metadata_logs.last)
					if notes_row_id.empty? # no internet connection
						@uploadButton.setEnabled(true)
						@not_yet_uploaded["notes"] = []
						@notes_metadata_logs.each do |log|
							hash = Hash.new
							log.each_pair { |key, value| hash[key] = value.to_s }
							@not_yet_uploaded["notes"] << hash
						end
						serialize_local_logs()
					else # sent to MySQL
						@notes_metadata_logs.last["id"] = notes_row_id
					end
				end
			end
		else
			puts "@can_add_logs = #{@can_add_logs}"
		end
	end
	ib_action :clickLogButton
	
	def clickUploadButton(sender)
		puts "clickUploadButton"
		#puts @metadata_logs.inspect
		if @can_add_logs
			not_uploaded_keywords = @metadata_logs.find_all { |m| m["id"]=="" && m["stop_time"]!=nil }
			not_uploaded_notes = @notes_metadata_logs.find_all { |m| m["id"]=="" && m["stop_time"]!=nil }
			upload_rows = not_uploaded_keywords + not_uploaded_notes
			unless upload_rows.empty?
				response = sendRowsToMySQL(upload_rows)
				alert = NSAlert.alloc.init
				if response.empty?
					alert.setMessageText("No rows were uploaded to the Lumberjack server")
					alert.setInformativeText("Please check your internet connection")
				else
					alert.setMessageText("#{response} rows were uploaded to the Lumberjack server")
					@not_yet_uploaded["locations"] = []
					@not_yet_uploaded["people"] = []
					@not_yet_uploaded["activities"] = []
					@not_yet_uploaded["others"] = []
					@not_yet_uploaded["notes"] = []
					serialize_local_logs()
					@uploadButton.setEnabled(false)
				end
				alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
			end
		else
			puts "@can_add_logs = #{@can_add_logs}"
		end
	end
	ib_action :clickUploadButton

	def applicationShouldTerminateAfterLastWindowClosed(sender)
		true
	end

	def applicationWillTerminate(sender)
		login(self)
	end

	### TableView methods ###

	# def control_textView_completions_forPartialWordRange_indexOfSelectedItem( control, textView, words, charRange, index) # press Esc for completion list
	#		if @tableView.editedColumn==1 # Keyword
	#			if @autocomplete_keywords.empty?
	#				return nil
	#			else
	#				return @autocomplete_keywords
	#			end
	#		else # Notes
	#			return nil
	#		end
	# end

	def control_textView_doCommandBySelector(control, textView, commandSelector)
		#puts "control:textView:doCommandBySelector:", control.tag, textView.inspect, commandSelector.inspect
		if commandSelector=="insertNewline:"
			#puts "RETURN was pressed"
			clickLogButton(self)
			return true
		elsif commandSelector=="insertTab:"
			#puts "TAB was pressed"
			return false
		else
			return false
		end
	end
	
	# def setRowHeights()
	# 	notesColumn = @notesTableView.tableColumns[2] # make cell once and reuse for all calcs
	# 	# puts "NSMakeRect?"
	# 	# rect = NSMakeRect(0.0, 0.0, notesColumn.width, 1000000.0)
	# 	cell = notesColumn.dataCellForRow(0)
	# 	cell.wraps = true
	# 	puts cell.wraps
	# 	@notes_metadata_logs.each do |row|
	# 		if row["height"].nil?
	# 			cell.attributedStringValue = row["metadata"]
	# 			size = cell.cellSize
	# 			puts size.inspect
	# 			# if rowHeight < 17.0
	# 				rowHeight = 17.0
	# 			# end
	# 			row["height"] = rowHeight
	# 		end
	# 	end
	# end
	#
	# def tableView_heightOfRow(tableView, rowIndex)
	# 	row = @notes_metadata_logs[rowIndex]
	# 	if row["height"].nil?
	# 		setRowHeights()
	# 	end
	# 	rowHeight = row["height"]
	# 	return rowHeight
	# end

	def numberOfRowsInTableView(tableView)
		tableColumn = tableView.tableColumns[0]
		columnName = tableColumn.headerCell.stringValue
		if columnName=="Locations" #tableView==@locationsTableView
			if @location_keywords.nil?
				return 0
			else
				return @location_keywords.length
			end
		elsif columnName=="People" #tableView==@peopleTableView
			if @people_keywords.nil?
				return 0
			else
				return @people_keywords.length
			end
		elsif columnName=="Activities" #tableView==@activitiesTableView
			if @activity_keywords.nil?
				return 0
			else
				return @activity_keywords.length
			end
		elsif columnName=="Other Keywords" #tableView==@othersTableView
			if @other_keywords.nil?
				return 0
			else
				return @other_keywords.length
			end
		else
			if @notes_metadata_logs.nil?
				return 0
			else
				return @notes_metadata_logs.length
			end
		end
	end

	def tableView_objectValueForTableColumn_row(tableView, tableColumn, rowIndex)
		columnName = tableColumn.headerCell.stringValue
		if columnName=="Locations" #tableView==@locationsTableView
			keyword = @location_keywords[rowIndex]["keyword"]
			tableColumn.dataCellForRow(rowIndex).setTitle(keyword)
			if @location_keywords[rowIndex][:checkbox].nil?
				@location_keywords[rowIndex][:checkbox] = 0
			end
			checkbox = @location_keywords[rowIndex][:checkbox]
			return checkbox
		elsif columnName=="People" #tableView==@peopleTableView
			keyword = @people_keywords[rowIndex]["keyword"]
			tableColumn.dataCellForRow(rowIndex).setTitle(keyword)
			if @people_keywords[rowIndex][:checkbox].nil?
				@people_keywords[rowIndex][:checkbox] = 0
			end
			checkbox = @people_keywords[rowIndex][:checkbox]
			return checkbox
		elsif columnName=="Activities" #tableView==@activitiesTableView
			keyword = @activity_keywords[rowIndex]["keyword"]
			tableColumn.dataCellForRow(rowIndex).setTitle(keyword)
			if @activity_keywords[rowIndex][:checkbox].nil?
				@activity_keywords[rowIndex][:checkbox] = 0
			end
			checkbox = @activity_keywords[rowIndex][:checkbox]
			return checkbox
		elsif columnName=="Other Keywords" #tableView==@othersTableView
			keyword = @other_keywords[rowIndex]["keyword"]
			tableColumn.dataCellForRow(rowIndex).setTitle(keyword)
			if @other_keywords[rowIndex][:checkbox].nil?
				@other_keywords[rowIndex][:checkbox] = 0
			end
			checkbox = @other_keywords[rowIndex][:checkbox]
			return checkbox
		else
			unless @notes_metadata_logs.nil?
				object = @notes_metadata_logs[rowIndex]
				case tableColumn.headerCell.stringValue
				when "Start Time"
					return object["local_start_time"] unless object.nil?
				when "Keyword"
					return object["keyword"] unless object.nil?
				when "Notes"
					return object["text"] unless object.nil?
				when "Stop Time"
					return object["local_stop_time"] unless object.nil?
				end
			end
		end
	end

	def tableView_shouldTrackCell_forTableColumn_row(tableView, cell, tableColumn, rowIndex) # triggered by clicking on the row
		puts "@can_add_logs = #{@can_add_logs}"
		if @can_add_logs==false
			return false
		end
		timeNow = NSDate.date.dateByAddingTimeInterval(@offsetSeconds)
		serverTimestamp = @serverTimestampFormatter.stringFromDate(timeNow)
		localizedTimeString = NSDateFormatter.localizedStringFromDate_dateStyle_timeStyle(timeNow, NSDateFormatterShortStyle, NSDateFormatterMediumStyle)
		columnName = tableColumn.headerCell.stringValue
		if columnName=="Locations" #tableView==@locationsTableView
			tableView.deselectRow(rowIndex)
			row = @location_keywords[rowIndex]
			keyword = row["keyword"]
			if row[:checkbox]==0 # Start Log: add a new row to @metadata_logs
				row_number = @metadata_logs.length
				story = @stories[@eventPopup.indexOfSelectedItem]
				new_log_row = { "id" => "", "story_id" => story["id"], "card" => @user_id, "start_time" => serverTimestamp, "local_start_time" => localizedTimeString, "metadata" => "l:#{keyword}", "keyword" => "#{keyword} - Location", "text" => "" }
				@metadata_logs << new_log_row
				row[:checkbox] = 1
			else # Stop Log: send to database, get row id number
				log_row = @metadata_logs.find { |m| m["metadata"]=="l:#{keyword}" && m["stop_time"].nil? }
				#puts "log_row = #{log_row.inspect}"
				if log_row.nil?
					#puts "NIL!!!!!"
				end
				log_row["stop_time"] = serverTimestamp
				log_row["local_stop_time"] = localizedTimeString
				row_inserted = sendRowToMySQL(log_row)
				if row_inserted.empty? # no internet connection
					@uploadButton.setEnabled(true)
					@not_yet_uploaded["locations"] = []
					@metadata_logs.each do |log|
						hash = Hash.new
						log.each_pair { |key, value| hash[key] = value.to_s }
						@not_yet_uploaded["locations"] << hash
					end
					serialize_local_logs()
				else # sent to MySQL
					log_row["id"] = @metadata_logs.count
				end
				row[:checkbox] = 0
			end
			return true
		elsif columnName=="People" #tableView==@peopleTableView
			tableView.deselectRow(rowIndex)
			row = @people_keywords[rowIndex]
			keyword = row["keyword"]
			if row[:checkbox]==0 # Start Log: add a new row to @metadata_logs
				row_number = @metadata_logs.length
				story = @stories[@eventPopup.indexOfSelectedItem]
				new_log_row = { "id" => "", "story_id" => story["id"], "card" => @user_id, "start_time" => serverTimestamp, "local_start_time" => localizedTimeString, "metadata" => "p:#{keyword}", "keyword" => "#{keyword} - Person", "text" => "" }
				@metadata_logs << new_log_row
				row[:checkbox] = 1
			else # Stop Log: send to database, get row id number
				log_row = @metadata_logs.find { |m| m["metadata"]=="p:#{keyword}" && m["stop_time"].nil? }
				puts "log_row = #{log_row.inspect}"
				if log_row.nil?
					puts "NIL!!!!!"
				end
				log_row["stop_time"] = serverTimestamp
				log_row["local_stop_time"] = localizedTimeString
				row_inserted = sendRowToMySQL(log_row)
				if row_inserted.empty? # no internet connection
					@uploadButton.setEnabled(true)
					@not_yet_uploaded["people"] = []
					@metadata_logs.each do |log|
						hash = Hash.new
						log.each_pair { |key, value| hash[key] = value.to_s }
						@not_yet_uploaded["people"] << hash
					end
					serialize_local_logs()
				else # sent to MySQL
					log_row["id"] = @metadata_logs.count
				end
				row[:checkbox] = 0
			end
			return true
		elsif columnName=="Activities" #tableView==@activitiesTableView
			tableView.deselectRow(rowIndex)
			row = @activity_keywords[rowIndex]
			keyword = row["keyword"]
			timeNow = NSDate.date.dateByAddingTimeInterval(@offsetSeconds)
			serverTimestamp = @serverTimestampFormatter.stringFromDate(timeNow)
			localizedTimeString = NSDateFormatter.localizedStringFromDate_dateStyle_timeStyle(timeNow, NSDateFormatterShortStyle, NSDateFormatterMediumStyle)
			if row[:checkbox]==0 # Start Log: add a new row to @metadata_logs
				row_number = @metadata_logs.length
				story = @stories[@eventPopup.indexOfSelectedItem]
				new_log_row = { "id" => "", "story_id" => story["id"], "card" => @user_id, "start_time" => serverTimestamp, "local_start_time" => localizedTimeString, "metadata" => "a:#{keyword}", "keyword" => "#{keyword} - Activity", "text" => "" }
				@metadata_logs << new_log_row
				row[:checkbox] = 1
			else # Stop Log: send to database, get row id number
				log_row = @metadata_logs.find { |m| m["metadata"]=="a:#{keyword}" && m["stop_time"].nil? }
				#puts "log_row = #{log_row.inspect}"
				if log_row.nil?
					#puts "NIL!!!!!"
				end
				log_row["stop_time"] = serverTimestamp
				log_row["local_stop_time"] = localizedTimeString
				row_inserted = sendRowToMySQL(log_row)
				if row_inserted.empty? # no internet connection
					@uploadButton.setEnabled(true)
					@not_yet_uploaded["activities"] = []
					@metadata_logs.each do |log|
						hash = Hash.new
						log.each_pair { |key, value| hash[key] = value.to_s }
						@not_yet_uploaded["activities"] << hash
					end
					serialize_local_logs()
				else # sent to MySQL
					log_row["id"] = @metadata_logs.count
				end
				row[:checkbox] = 0
			end
			return true
		elsif columnName=="Other Keywords" #tableView==@othersTableView
			tableView.deselectRow(rowIndex)
			row = @other_keywords[rowIndex]
			keyword = row["keyword"]
			if row[:checkbox]==0 # Start Log: add a new row to @metadata_logs
				row_number = @metadata_logs.length
				story = @stories[@eventPopup.indexOfSelectedItem]
				new_log_row = { "id" => "", "story_id" => story["id"], "card" => @user_id, "start_time" => serverTimestamp, "local_start_time" => localizedTimeString, "metadata" => "k:#{keyword}", "keyword" => "#{keyword} - Other", "text" => "" }
				@metadata_logs << new_log_row
				row[:checkbox] = 1
			else # Stop Log: send to database, get row id number
				log_row = @metadata_logs.find { |m| m["metadata"]=="k:#{keyword}" && m["stop_time"].nil? }
				#puts "log_row = #{log_row.inspect}"
				if log_row.nil?
					#puts "NIL!!!!!"
				end
				log_row["stop_time"] = serverTimestamp
				log_row["local_stop_time"] = localizedTimeString
				row_inserted = sendRowToMySQL(log_row)
				if row_inserted.empty? # no internet connection
					@uploadButton.setEnabled(true)
					@not_yet_uploaded["others"] = []
					@metadata_logs.each do |log|
						hash = Hash.new
						log.each_pair { |key, value| hash[key] = value.to_s }
						@not_yet_uploaded["others"] << hash
					end
					serialize_local_logs()
				else # sent to MySQL
					log_row["id"] = @metadata_logs.count
				end
				row[:checkbox] = 0
			end
			return true
		else # Notes tableView
			return false
		end
	end

	def tableView_setObjectValue_forTableColumn_row(tableView, anObject, tableColumn, rowIndex)
		unless @notes_metadata_logs[rowIndex].nil?
			case tableColumn.headerCell.stringValue
			when "Keyword"
				@notes_metadata_logs[rowIndex]["metadata"] = "#{@keyword_prefix}#{anObject}"
				case @keyword_prefix
				when "p:"
					@notes_metadata_logs[rowIndex]["keyword"] = "#{anObject} - Person"
				when "a:"
					@notes_metadata_logs[rowIndex]["keyword"] = "#{anObject} - Activity"
				when "k:"
					@notes_metadata_logs[rowIndex]["keyword"] = "#{anObject} - Other"
				else
					@notes_metadata_logs[rowIndex]["keyword"] = "#{anObject}"
				end
				@logged_keywords << anObject.to_s
				rebuildComboBox()
				# TODO: maybe able to edit a row instead of insert if the API knows the id number of the row
			when "Notes"
				@notes_metadata_logs[rowIndex]["text"] = anObject.to_s
			end
		end
	end

	#	def tableViewSelectionDidChange(notification) # this causes spinning beachball on High Sierra
	#		#puts "tableViewSelectionDidChange"
	#		selected_row_index = notification.object.selectedRowIndexes.to_a.first
	#		if selected_row_index.nil?
	#			#puts "no row is selected"
	#		else
	#			selected_row = @notes_metadata_logs[selected_row_index]
	#		end
	#	end

	# def tableView_sortDescriptorsDidChange(tableView, oldDescriptors)
	# 	puts "tableView_sortDescriptorsDidChange"
	# 	unless @notes_metadata_logs.nil?
	# 		sortDescriptor = @notesTableView.sortDescriptors.first
	# 		puts sortDescriptor
	# 		key = sortDescriptor.key
	# 		ascending = sortDescriptor.ascending
	# 		sorted_logs = @notes_metadata_logs.sort_by { |m| m["start_time"] } #.to_ns.sortedArrayUsingDescriptors([sortDescriptor])
	# 		if ascending==true
	# 			@notes_metadata_logs = sorted_logs
	# 		else
	# 			@notes_metadata_logs = sorted_logs.reverse
	# 		end
	# 		@notesTableView.reloadData
	# 	end
	# end
	
	def serialize_local_logs
		#puts "serialize_local_logs", @local_logs_db
		# File.open(@marshal_will_and_holly, 'w') { |f| Marshal.dump(@local_logs_db, f) }
		is_valid = NSJSONSerialization.isValidJSONObject(@local_logs_db)
		if is_valid
			file = NSOutputStream.alloc.initToFileAtPath_append(@argonaut, false)
			file.open
			NSJSONSerialization.writeJSONObject_toStream_options_error(@local_logs_db, file, 0, nil)
			file.close
		end
	end
	
end

